/*-
 * Copyright (c) 2004 Pawel Jakub Dawidek <pjd@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/linker.h>
#include <sys/module.h>
#include <sys/disk.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <limits.h>
#include <paths.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <assert.h>
#include <libgeom.h>
#include <geom/null/g_null.h>


enum {
	UNSET,
	CREATE,
	DESTROY
} action = UNSET;

static const char *comm;
static int force = 0;
static int verbose = 0;


static void
usage(void)
{
	fprintf(stderr,
	    "usage: %s create [-v] provider1 [provider2 [...]]\n"
	    "       %s destroy [-fv] provider1 [provider2 [...]]\n",
	    comm, comm);
	exit(EXIT_FAILURE);
}

static void
load_module(void)
{

	if (modfind("g_null") < 0) {
		/* Not present in kernel, try loading it. */
		if (kldload("geom_null") < 0 || modfind("g_null") < 0) {
			if (errno != EEXIST) {
				errx(EXIT_FAILURE,
				    "geom_null module not available!");
			}
		}
	}
}

static struct gctl_req *
init_gctl(const char *name, const char *verb)
{
	struct gctl_req *grq;

	grq = gctl_get_handle();
	gctl_ro_param(grq, "verb", -1, verb);
	gctl_ro_param(grq, "class", -1, G_NULL_CLASS_NAME);
	if (name != NULL)
		gctl_ro_param(grq, "geom", -1, name);

	return (grq);
}

static int
null_create(const char *provider)
{
	struct gctl_req *grq;
	const char *errstr;

	load_module();

	if (verbose)
		printf("Creating device %s%s...\n", provider, G_NULL_SUFFIX);

	grq = init_gctl(NULL, "create device");
	gctl_ro_param(grq, "provider", -1, provider);
	errstr = gctl_issue(grq);
	if (errstr == NULL) {		
		gctl_free(grq);
		if (verbose)
			printf("Done.\n");
		return (0);
	}

	fprintf(stderr, "%s\n", errstr);
	gctl_free(grq);
	return (1);
}

static int
null_destroy(const char *provider)
{
	struct gctl_req *grq;
	const char *errstr;

	if (verbose)
		printf("Destroying device %s...\n", provider);

	grq = init_gctl(provider, "destroy device");
	gctl_ro_param(grq, "force", sizeof(force), &force);
	errstr = gctl_issue(grq);
	if (errstr == NULL) {		
		gctl_free(grq);
		if (verbose)
			printf("Done.\n");
		return (0);
	}

	fprintf(stderr, "%s\n", errstr);
	gctl_free(grq);
	return (1);
}

int
main(int argc, char *argv[])
{
	int (*func)(const char *);
	int ch, exitcode, i;

	comm = basename(argv[0]);

	if (argc < 2)
		usage();

	if (strcasecmp(argv[1], "create") == 0)
		action = CREATE;
	else if (strcasecmp(argv[1], "destroy") == 0)
		action = DESTROY;
	else
		usage();
	argv++;
	argc--;

	while ((ch = getopt(argc, argv, "fv")) != -1) {
		switch (ch) {
		case 'f':
			if (action != DESTROY)
				usage();
			force = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 0)
		usage();

	func = NULL;	/* Shut up compiler. */
	switch (action) {
	case CREATE:
		func = null_create;
		break;
	case DESTROY:
		func = null_destroy;
		break;
	case UNSET:
	default:
		usage();
		/* NOTREACHED */
	}

	exitcode = EXIT_SUCCESS;
	for (i = 0; i < argc; i++) {
		if (func(argv[i]) != 0)
			exitcode = EXIT_FAILURE;
	}

	exit(exitcode);
}
